from copy import deepcopy

K8S_SCALE = {
    "kind": "Scale",
    "metadata": {
        "name": "",
        "namespace": ""
    },
    "spec": {
        "replicas": 0
    }
}


class K8SScale(object):

    def __init__(
            self, namespace, name, scale):
        self._raw = self.create(
            namespace,
            name=name,
            scale=scale)

    @property
    def id(self):
        return self._raw['metadata']['name']

    def create(self, namespace, name, scale):
        scale_info = deepcopy(K8S_SCALE)
        scale_info['metadata']['name'] = name
        scale_info['metadata']['namespace'] = namespace
        scale_info['spec']['replicas'] = scale
        return scale_info

    def payload(self):
        return deepcopy(self._raw)



from copy import deepcopy

K8S_DELETE = {
    'kind': 'DeleteOptions',
    'propagationPolicy': 'Background'
}


class K8SDelete(object):

    def __init__(self, purge=True):
        self._raw = K8S_DELETE
        # K8S supports multiple ways of dealing with the pods but
        # Foreground does not delete the job until pods are deleted
        # This is a different approach from nomad.
        # https://github.com/kubernetes/kubernetes/blob/release-1.7/staging/src/k8s.io/apimachinery/pkg/apis/meta/v1/types.go#L360-L370  # noqa
        # if purge:
        # 	self._raw['propagationPolicy'] = 'Foreground'

    def payload(self):
        return deepcopy(self._raw)



from aioclustermanager.k8s.caller import K8SCaller
from base64 import b64decode

import aiohttp
import asyncio
import logging
import os
import ssl
import tempfile

logger = logging.getLogger("aioclustermanager")

SERVICE_HOST_ENV_NAME = "KUBERNETES_SERVICE_HOST"
SERVICE_PORT_ENV_NAME = "KUBERNETES_SERVICE_PORT"
SERVICE_TOKEN_ENV_NAME = "KUBERNETES_SERVICE_TOKEN"
SERVICE_TOKEN_FILENAME = "/var/run/secrets/kubernetes.io/serviceaccount/token"
SERVICE_CERT_FILENAME = "/var/run/secrets/kubernetes.io/serviceaccount/ca.crt"


def _join_host_port(host, port):
    template = "%s:%s"
    host_requires_bracketing = ":" in host or "%" in host
    if host_requires_bracketing:
        template = "[%s]:%s"
    return template % (host, port)


class Configuration:
    file = None
    ssl_context = None
    cert_file = None
    scheme = "https"

    def __init__(self, environment, loop=None):
        self.headers = {}
        self.ssl_context = None
        self.basic_auth = None
        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop

        self.environment = environment

        # Authentication
        auth = self.environment.get("auth", "in_cluster")
        if auth == "in_cluster":
            self.load_in_cluster()
        elif auth == "certificate":
            self.load_certificate()
        elif auth == "certificate_file":
            self.load_certificate_file()
        elif auth == "basic_auth":
            self.basic_auth = aiohttp.BasicAuth(self.environment["user"], self.environment["credentials"])
        elif auth == "token":
            self.headers = {"Authorization": "Bearer " + b64decode(self.environment["token"]).decode("utf-8")}

        # We create the aiohttp client session
        if self.environment.get("skip_ssl", "false").lower() == "false":
            self.ssl_context = False
        else:
            if self.ssl_context is None:
                self.ssl_context = ssl.SSLContext()

            if self.environment.get("ca") is not None:
                self.ssl_context.load_verify_locations(cadata=b64decode(self.environment["ca"]))
            elif self.environment.get("ca_file") is not None:
                self.ssl_context = ssl.SSLContext()
                self.ssl_context.load_verify_locations(cafile=self.environment["ca_file"])

        if "http_scheme" in environment:
            self.scheme = self.environment["http_scheme"]

        conn = aiohttp.TCPConnector(ssl=self.ssl_context, loop=self.loop)
        self.session = aiohttp.ClientSession(connector=conn, headers=self.headers, loop=self.loop)

    def load_in_cluster(self):
        # If the pod is running in cluster
        if SERVICE_TOKEN_ENV_NAME in os.environ:
            token = os.environ[SERVICE_TOKEN_ENV_NAME]
        else:
            with open(SERVICE_TOKEN_FILENAME) as fi:
                token = fi.read()
        self.headers = {"Authorization": "Bearer " + token}
        self.environment.update(
            {
                "ca_file": SERVICE_CERT_FILENAME,
                "endpoint": _join_host_port(
                    os.environ[SERVICE_HOST_ENV_NAME], os.environ[SERVICE_PORT_ENV_NAME]
                ),
            }
        )

    def load_certificate_file(self):
        self.ssl_context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
        if "key" in self.environment:
            self.ssl_context.load_cert_chain(
                certfile=self.environment["certificate"], keyfile=self.environment["key"]
            )
        else:
            self.ssl_context.load_cert_chain(certfile=self.environment["certificate"])

    def load_certificate(self):
        self.ssl_context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH)
        self.cert_file = tempfile.NamedTemporaryFile(delete=False)
        self.cert_file.write(b64decode(self.environment["certificate"]))
        self.cert_file.close()
        self.client_key = tempfile.NamedTemporaryFile(delete=False)
        self.client_key.write(b64decode(self.environment["key"]))
        self.client_key.close()
        self.ssl_context.load_cert_chain(certfile=self.cert_file.name, keyfile=self.client_key.name)


class K8SContextManager:
    def __init__(self, environment, loop=None):
        self.environment = environment
        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop

    async def __aenter__(self):
        return await self.open()

    async def open(self):
        self.config = Configuration(self.environment, self.loop)
        return K8SCaller(
            self.config.ssl_context,
            self.environment["endpoint"],
            self.config.session,
            scheme=self.config.scheme,
        )

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def close(self):
        if self.config.file is not None:
            os.unlink(self.config.file.name)
        await self.config.session.close()


async def create_k8s_caller(environment):
    config = Configuration(environment)
    return K8SCaller(config.ssl_context, environment["endpoint"], config.session, scheme=config.scheme)



from pathlib import Path

import os
import yaml


def get_inner_config(list_info, key):
    for element in list_info:
        if element["name"] == key:
            return element


def get_k8s_config():
    home = str(Path.home())
    # Designed to run on docker desktop k8s support
    # XXX DOES NOT CURRENTLY WORK WITH containers running in k8s already
    if os.environ.get("TEST_INSIDE_K8S", False):
        config_k8s = {"auth": "in_cluster"}
        return config_k8s

    with open(home + "/.kube/config", "r") as f:
        configuration = yaml.load(f)

    TRAVIS = os.environ.get("TRAVIS", "false")
    config_k8s = {}
    # if TRAVIS == "true":
    #     # defined on testing.csv
    #     config_k8s["user"] = "testinguser"
    #     config_k8s["credentials"] = "12345678"
    #     config_k8s["auth"] = "basic_auth"
    #     cluster_info = get_inner_config(configuration["clusters"], "minikube")
    #     server_url = cluster_info["cluster"]["server"]
    #     if "://" in server_url:
    #         schema, server_url = server_url.split("://")

    #     config_k8s["endpoint"] = server_url
    #     config_k8s["http_scheme"] = schema
    #     config_k8s["ca_file"] = cluster_info["cluster"]["certificate-authority"]
    # else:
    local_cluster = None
    for context in configuration["contexts"]:
        if context["name"] == "docker-desktop":
            local_cluster = "docker-desktop"
            break
        elif context["name"] == "minikube":
            local_cluster = "minikube"
            break

    if local_cluster is None:
        raise Exception("Invalid cluster configuration")

    user_info = get_inner_config(configuration["users"], local_cluster)
    cluster_info = get_inner_config(configuration["clusters"], local_cluster)

    if TRAVIS == "true":
        config_k8s["token"] = os.environ["K8S_TOKEN"]
        config_k8s["auth"] = "token"
    elif local_cluster == "minikube":
        config_k8s["certificate"] = user_info["user"]["client-certificate"]
        config_k8s["key"] = user_info["user"]["client-key"]
        config_k8s["auth"] = "certificate_file"
    elif local_cluster == "docker-desktop":
        config_k8s["certificate"] = user_info["user"]["client-certificate-data"]
        config_k8s["key"] = user_info["user"]["client-key-data"]
        config_k8s["auth"] = "certificate"

    server_url = cluster_info["cluster"]["server"]
    schema = "http"
    if "://" in server_url:
        schema, server_url = server_url.split("://")
    config_k8s["endpoint"] = server_url
    config_k8s["http_scheme"] = schema
    config_k8s["skip_ssl"] = (
        "true" if cluster_info["cluster"].get("insecure-skip-tls-verify", False) else "false"
    )
    if config_k8s["skip_ssl"] == "false":
        config_k8s["ca"] = cluster_info["cluster"]["certificate-authority"]

    return config_k8s



import pytest
import asyncio

pytestmark = pytest.mark.asyncio


async def test_get_jobs_k8s(kubernetes):
    # We clean up all the jobs on the namespace

    result = await kubernetes.get_nodes()
    assert len(result) > 0

    result = await kubernetes.cleanup_jobs("aiocluster-test")
    assert result == 0

    jobs_info = await kubernetes.list_jobs("aiocluster-test")
    assert jobs_info.total == 0

    await kubernetes.create_job(
        "aiocluster-test",  # namespace
        "test-job",  # jobid
        "perl",  # image
        command=["perl", "-Mbignum=bpi", "-wle", "print bpi(2000)"],
    )

    job_info = await kubernetes.get_job("aiocluster-test", "test-job")
    assert job_info.id == "test-job"

    jobs_info = await kubernetes.list_jobs("aiocluster-test")
    assert jobs_info.total == 1

    result = await kubernetes.wait_for_job("aiocluster-test", "test-job")
    assert result == 1

    job_info = await kubernetes.get_job("aiocluster-test", "test-job")
    assert job_info.finished
    assert job_info.id == "test-job"

    executions = await kubernetes.list_job_executions("aiocluster-test", "test-job")
    assert len(executions) > 0

    log = await kubernetes.get_execution_log("aiocluster-test", "test-job", executions[0].internal_id)
    assert "3.14" in log

    result = await kubernetes.delete_job("aiocluster-test", "test-job")
    assert result is True

    job_info = await kubernetes.get_job("aiocluster-test", "test-job")
    assert job_info is None


async def test_get_jobs_limit_k8s(kubernetes):
    # We clean up all the jobs on the namespace

    result = await kubernetes.define_quota("aiocluster-test", cpu_limit="400m", mem_limit="900M")
    assert result is True

    result = await kubernetes.create_job(
        "aiocluster-test",  # namespace
        "test-job",  # jobid
        "tensorflow/tensorflow:1.13.2-py3-jupyter",  # image
        cpu_limit="300m",
        mem_limit="800M",
        command=["jupyter-nbconvert"],
        args=[
            "--execute",
            "/tf/tensorflow-tutorials/basic_classification.ipynb",
            "--ExecutePreprocessor.timeout=380",
        ],
    )
    assert result is True

    await asyncio.sleep(5)

    result = await kubernetes.create_job(
        "aiocluster-test",  # namespace
        "test-job-2",  # jobid
        "tensorflow/tensorflow:1.13.2-py3-jupyter",  # image
        cpu_limit="300m",
        mem_limit="800M",
        command=["jupyter-nbconvert"],
        args=[
            "--execute",
            "/tf/tensorflow-tutorials/basic_classification.ipynb",
            "--ExecutePreprocessor.timeout=380",
        ],
    )
    assert result is True

    await asyncio.sleep(20)

    # we want to wait that the first job starts

    result = await kubernetes.wait_for_job_execution_status("aiocluster-test", "test-job")
    assert result == "Running"

    result = await kubernetes.waiting_jobs("aiocluster-test")
    assert len(result) == 1
    assert result[0] == "test-job-2"

    result = await kubernetes.running_jobs("aiocluster-test")
    assert len(result) == 1
    assert result[0] == "test-job"

    result = await kubernetes.delete_job("aiocluster-test", "test-job")
    assert result is True

    await asyncio.sleep(50)

    result = await kubernetes.waiting_jobs("aiocluster-test")
    assert len(result) >= 0

    result = await kubernetes.running_jobs("aiocluster-test")
    assert len(result) <= 1


# async def test_get_tfjobs_k8s(kubernetes):
#     # We clean up all the jobs on the namespace

#     result = await kubernetes.define_quota("aiocluster-test", cpu_limit="400m", mem_limit="900M")
#     assert result is True

#     await kubernetes.install_tfjobs("aiocluster-test")

#     result = await kubernetes.create_tfjob(
#         "aiocluster-test",  # namespace
#         "test-tfjob",  # jobid
#         "gcr.io/tf-on-k8s-dogfood/tf_sample:dc944ff",  # image
#         cpu_limit="300m",
#         mem_limit="800M",
#         workers=1,
#         ps=2,
#         masters=1,
#     )
#     assert result is True

#     await kubernetes.get_tfjob("aiocluster-test", "test-tfjob")

#     await kubernetes.list_tfjobs("aiocluster-test")



from aioclustermanager.nomad import const
from aioclustermanager.nomad.executions_list import NomadExecutionList
from aioclustermanager.nomad.job import NomadJob
from aioclustermanager.nomad.job_list import NomadJobList
from aioclustermanager.nomad.namespace import NomadNamespace
from aioclustermanager.nomad.node_list import NomadNodeList

import asyncio
import logging

# from aioclustermanager.utils import convert

logger = logging.getLogger(__name__)

WATCH_OPS = {
    'job': 'http://{endpoint}/v1/job/{namespace}-{name}',
    'execution': 'http://{endpoint}/v1/job/{namespace}-{name}/allocations/{execution}',  # noqa
    'namespace': None
}

GET_OPS = {
    'list_jobs': 'http://{endpoint}/v1/jobs',
    'job': 'http://{endpoint}/v1/job/{namespace}-{name}',
    'executions': 'http://{endpoint}/v1/job/{namespace}-{name}/allocations',
    'log': 'http://{endpoint}/v1/client/fs/logs/{name}',
    'nodes': 'http://{endpoint}/v1/nodes'
}

POST_OPS = {
    'jobs': 'http://{endpoint}/v1/jobs',
    'job': 'http://{endpoint}/v1/job/{namespace}-{name}'
}

DELETE_OPS = {
    'job': 'http://{endpoint}/v1/job/{namespace}-{name}?purge={purge}',
    'execution': 'http://{endpoint}/v1/job/{namespace}-{name}/allocations/{execution}?purge={purge}',  # noqa
}


class NomadCaller:
    _type = 'nomad'
    constants = const

    def __init__(self, endpoint, datacenters, session):
        self.endpoint = endpoint
        self.session = session
        self.datacenters = datacenters

    @property
    def type(self):
        return self._type

    # Namespaces, implementation using prefix jobs

    async def get_namespace(self, namespace):
        return NomadNamespace(name=namespace)

    async def create_namespace(self, namespace):
        return True

    async def delete_namespace(self, namespace):
        jobs = await self.list_jobs(namespace)
        for job in jobs:
            await self.delete_job(namespace, job.id)
        return True

    async def get_nodes(self):
        url = GET_OPS['nodes']
        url = url.format(
            endpoint=self.endpoint)
        result = await self.get(url, {})
        if result is None:
            return None
        else:
            return NomadNodeList(data=result)

    async def wait_added(self, kind, namespace, name=None, timeout=30):
        url = WATCH_OPS[kind]
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        return await self.watch(url, not_value=['pending'], timeout=timeout)

    async def wait_deleted(self, kind, namespace, name=None, execution_id=None):  # noqa
        url = WATCH_OPS[kind]
        if url is None:
            return False
        url = url.format(
            namespace=namespace,
            name=name,
            execution=execution_id,
            endpoint=self.endpoint)
        return await self.watch(url, value=['dead', 'complete'])

    async def define_quota(self, namespace, cpu_limit=None, mem_limit=None):
        # Only supported on nomad enterprise
        pass

    async def get_job(self, namespace, name):
        url = GET_OPS['job']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        params = {}
        result = await self.get(url, params)
        if result is None:
            return None
        else:
            job_obj = NomadJob(data=result)
            job_obj._namespace = namespace
            return job_obj

    async def get_tfjob(self, namespace, name):
        # TODO
        return None

    async def get_scale_deploy(self, namespace, name):
        job = await self.get_job(namespace, name)
        return job.scale

    async def set_scale_deploy(self, namespace, name, scale):
        job = await self.get_job(namespace, name)
        job.rewrap()
        url = POST_OPS['jobs']
        job.scale = scale
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        job._namespace = namespace
        return await self.post(url, None, job.payload())

    async def get_job_executions(self, namespace, name):
        url = GET_OPS['executions']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        params = {}
        result = await self.get(url, params)
        if result is None:
            return None
        else:
            return NomadExecutionList(data=result)

    async def get_execution_log(self, namespace, job_id, execution_id):
        url = GET_OPS['log']
        url = url.format(
            namespace=namespace,
            name=execution_id,
            endpoint=self.endpoint)
        params = {
            'plain': 'true',
            'type': 'stdout',
            'task': namespace + '-' + job_id
        }
        result = await self.get(url, params)
        if result is None:
            return None
        else:
            return result

    async def get_execution_log_watch(self, namespace, job_id, execution_id):
        url = GET_OPS['log']
        url = url.format(
            namespace=namespace,
            name=execution_id,
            endpoint=self.endpoint)
        params = {
            'plain': 'true',
            'type': 'stdout',
            'task': namespace + '-' + job_id,
            'follow': 'true'
        }
        async for logline in await self._watch_log(url, params, timeout=3660):
            yield logline

    async def delete_job(self, namespace, name, wait=False, purge=True):
        url = DELETE_OPS['job']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint,
            purge='true' if purge else 'false')
        await self.delete(url, None, None)
        if wait:
            return await self.wait_deleted('job', namespace, name)
        return True

    async def delete_execution(self, namespace, job_id,
                               execution_id, wait=False, purge=True):
        url = DELETE_OPS['execution']
        url = url.format(
            namespace=namespace,
            name=job_id,
            execution=execution_id,
            endpoint=self.endpoint,
            purge='true' if purge else 'false')
        await self.delete(url, None, None)
        if wait:
            return await self.wait_deleted(
                'execution', namespace, job_id, execution_id)
        return True

    async def delete_tfjob(self, namespace, name, wait=False):
        # TODO
        return None

    async def list_jobs(self, namespace):
        url = GET_OPS['list_jobs']
        params = {
            'prefix': namespace + '-'
        }
        url = url.format(
            namespace=namespace,
            endpoint=self.endpoint)
        result = await self.get(url, params)
        if result is None:
            return None
        else:
            return NomadJobList(data=result, namespace=namespace)

    async def list_tfjobs(self, namespace):
        # TODO
        return []

    async def create_job(
            self, namespace, name, image,
            command=None, args=None,
            cpu_limit=None, mem_limit=None,
            envvars={}, volumes=None, volumeMounts=None,
            envFrom=None, entrypoint=None, **kw):
        url = POST_OPS['jobs']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        obj = NomadJob(
            namespace=namespace,
            name=name,
            image=image,
            command=command, args=args,
            cpu_limit=cpu_limit, mem_limit=mem_limit,
            envvars=envvars, volumes=volumes, volumeMounts=volumeMounts,
            envFrom=envFrom, entrypoint=entrypoint, **kw)
        obj._namespace = namespace
        obj.set_datacenters([self.datacenters])
        return await self.post(url, None, obj.payload())

    async def create_tfjob(
            self, namespace, name, image,
            command=None, args=None,
            cpu_limit=None, mem_limit=None,
            envvars={}, workers=1, ps=1,
            masters=1, tb_gs=None, **kw):
        # TODO
        return None

    async def wait_for_job(self, namespace, name):
        url = WATCH_OPS['job']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        await self.watch(url, value=['dead', 'complete'])
        job = await self.get_job(namespace, name)
        if not job.finished:
            raise Exception('Not finished')
        executions = await self.get_job_executions(namespace, name)
        return 1 if executions.is_done() else 0

    # BASIC OPS

    async def _watch_log(self, url, params, timeout=20):
        async with self.session.get(
                url,
                params=params,
                timeout=timeout) as resp:
            assert resp.status == 200
            while True:
                data = await resp.content.readline()
                yield data

    async def watch(self, url, value=None, not_value=None, timeout=30):
        return await asyncio.wait_for(
            self._watch(url, value, not_value), timeout)

    async def _watch(self, url, value=None, not_value=None):
        if value is not None and not isinstance(value, list):
            value = [value]
        if not_value is not None and not isinstance(not_value, list):
            not_value = [not_value]

        not_found = True
        while not_found:
            job = await self.get(url, {})
            if job is not None:
                if not_value is not None:
                    not_found = job['Status'] in not_value
                elif value is not None:
                    not_found = job['Status'] not in value
            if not_found is True:
                await asyncio.sleep(1)
        return job

    async def get(self, url, params):
        async with self.session.get(
                url,
                headers={
                    'Accept': 'application/json'
                },
                params=params) as resp:
            if resp.status == 404:
                return None
            assert resp.status == 200
            result = await resp.json()
            return result

    async def post(self, url, version, payload):
        async with self.session.post(
                url,
                json=payload,
                headers={
                    'Content-Type': 'application/json',
                    'Accept': 'application/json'
                }) as resp:
            if resp.status in [201, 200]:
                return await resp.json()
            else:
                text = await resp.text()
                raise Exception('Error call: %d - %s' % (resp.status, text))

    async def delete(self, url, version, payload):
        async with self.session.delete(
                url,
                headers={
                    'Content-Type': 'application/json',
                    'Accept': 'application/json'
                }) as resp:
            if resp.status == 200:
                return await resp.json()
            elif resp.status == 404:
                return None
            else:
                text = await resp.text()
                raise Exception('Error call: %d - %s' % (resp.status, text))

    async def get_config_maps(self, namespace, labels=None):
        raise NotImplementedError()



from aioclustermanager.job import Job
from aioclustermanager.nomad.const import DEAD, SUCCEEDED
from copy import deepcopy

import json

NOMAD_JOB = {
    "Job": {
        "Constraints": [],
        "ID": "",
        "Name": "",
        "TaskGroups": [{
            "Constraints": None,
            "Name": "aioclustermanager",
            "EphemeralDisk": {
                "Migrate": False,
                "SizeMB": 300,
                "Sticky": False
            },
            "RestartPolicy": {
                "Attempts": 0,
                "Delay": 15000000000,
                "Interval": 60000000000,
                "Mode": "fail"
            },
            "Tasks": [{
                "Config": {
                    "image": ""
                },
                "Constraints": [],
                "Driver": "docker",
                "Env": {},
                "LogConfig": {
                    "MaxFileSizeMB": 10,
                    "MaxFiles": 10
                },
                "Name": "",
                "Resources": {
                    "CPU": 200,
                    "MemoryMB": 300,
                    "Networks": [{
                        "CIDR": "",
                        "Device": "",
                        "DynamicPorts": [{
                            "Label": "db",
                            "Value": 0
                        }],
                        "IP": "",
                        "MBits": 200,
                        "ReservedPorts": None
                    }]
                },
                "Services": [],
                "Templates": [],
                "User": "",
            }]
        }],
        "Type": "batch",
    }
}


class NomadJob(Job):

    @property
    def namespace(self):
        return getattr(self, '_namespace', '')

    @property
    def active(self):
        return self._raw['Status'] not in (DEAD, SUCCEEDED)

    @property
    def finished(self):
        return self._raw['Status'] in (DEAD, SUCCEEDED)

    @property
    def status(self):
        return self._raw['Status']

    @property
    def id(self):
        if self.namespace == '':
            return self._raw['ID']
        elif self._raw['ID'].startswith(self.namespace):
            to_remove = len(self.namespace) + 1
            return self._raw['ID'][to_remove:]

    @property
    def command(self):
        return self._raw['Job']['TaskGroups'][0]['Tasks'][0]['Config']['command']  # noqa

    @property
    def image(self):
        return self._raw['Job']['TaskGroups'][0]['Tasks'][0]['Config']['image']  # noqa

    @property
    def scale(self):
        if 'TaskGroups' in self._raw:
            return self._raw['TaskGroups'][0]['Count']
        else:
            return self._raw['Job']['TaskGroups'][0]['Count']

    @scale.setter
    def scale(self, scale):
        if 'TaskGroups' in self._raw:
            self._raw['TaskGroups'][0]['Count'] = scale
        else:
            self._raw['Job']['TaskGroups'][0]['Count'] = scale

    def rewrap(self):
        self._raw = {
            'Job': self._raw
        }

    def get_payload(self):
        if 'TaskGroups' not in self._raw:
            return
        try:
            return json.loads(
                self._raw['TaskGroups'][0]['Tasks'][0]['Env']['PAYLOAD'])
        except (TypeError, KeyError, json.decoder.JSONDecodeError):
            try:
                task = self._raw['TaskGroups'][0]['Tasks'][0]
                templates = task['Templates']
                template = templates[-1]['EmbeddedTmpl']
                payload = template.split(
                    'cat <<EOF')[-1].split('EOF')[0].strip()
                return json.loads(payload)
            except (TypeError, AttributeError, KeyError,
                    json.decoder.JSONDecodeError):
                pass

    def set_datacenters(self, datacenters):
        self._raw['Job']['Datacenters'] = datacenters

    def create(self, namespace, name, image, **kw):
        job_info = deepcopy(NOMAD_JOB)

        real_job_id = '{}-{}'.format(namespace, name)

        job_info['Job']['ID'] = real_job_id
        job_info['Job']['Name'] = real_job_id
        job_info['Job']['TaskGroups'][0]['Tasks'][0]['Name'] = real_job_id

        job_info['Job']['TaskGroups'][0]['Tasks'][0]['Config']['image'] = image

        if 'nomad_constraints' in kw and kw['nomad_constraints'] is not None:
            job_info['Job']['Constraints'] = kw['nomad_constraints']

        if 'docker_network_mode' in kw and kw['docker_network_mode'] is not None:  # noqa
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Config']['network_mode'] = kw['docker_network_mode']  # noqa

        if 'volumes' in kw and kw['volumes'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Config']['volumes'] = kw['volumes']  # noqa

        extend_args = []
        if 'command' in kw and kw['command'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Config']['command'] = kw['command'][0]  # noqa

            if len(kw['command']) > 1:
                extend_args = kw['command'][1:]

        if 'args' in kw and kw['args'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Config']['args'] = extend_args + kw['args']  # noqa

        if 'templates' in kw and kw['templates'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Templates'] = kw['templates']  # noqa

        if 'cpu_limit' in kw and kw['cpu_limit'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Resources']['CPU'] = kw['cpu_limit']  # noqa

        if 'mem_limit' in kw and kw['mem_limit'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Resources']['MemoryMB'] = kw['mem_limit']  # noqa

        if 'envvars' in kw and kw['envvars'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Env'] = kw['envvars']  # noqa

        if 'datacenters' in kw and kw['datacenters'] is not None:
            job_info['Job']['Datacenters'] = kw['datacenters']

        return job_info



from aioclustermanager.k8s import const
from aioclustermanager.k8s.delete import K8SDelete
from aioclustermanager.k8s.executions_list import K8SExecutionList
from aioclustermanager.k8s.job import K8SJob
from aioclustermanager.k8s.job_list import K8SJobList
from aioclustermanager.k8s.namespace import K8SNamespace
from aioclustermanager.k8s.node_list import K8SNodeList
from aioclustermanager.k8s.quota import K8SQuota
from aioclustermanager.k8s.tf_job import K8STFJob
from aioclustermanager.k8s.tf_job_list import K8STFJobList
from aioclustermanager.k8s.scale import K8SScale

import concurrent
import json
import logging

logger = logging.getLogger(__name__)

WATCH_OPS = {
    "namespace": "{scheme}://{endpoint}/api/v1/watch/namespaces/{namespace}",
    "job": "{scheme}://{endpoint}/apis/batch/v1/watch/namespaces/{namespace}/jobs/{name}",  # noqa
    "execution": "{scheme}://{endpoint}/api/v1/watch/namespaces/{namespace}/pods/{name}",  # noqa
    "tfjob": "{scheme}://{endpoint}/apis/kubeflow.org/v1alpha1/watch/namespaces/{namespace}/tfjobs/{name}",  # noqa
}

GET_OPS = {
    "namespace": "{scheme}://{endpoint}/api/v1/namespaces/{namespace}",
    "list_jobs": "{scheme}://{endpoint}/apis/batch/v1/namespaces/{namespace}/jobs",
    "job": "{scheme}://{endpoint}/apis/batch/v1/namespaces/{namespace}/jobs/{name}/status",  # noqa
    "executions": "{scheme}://{endpoint}/api/v1/namespaces/{namespace}/pods/?labelSelector=job-name={name}",  # noqa
    "tfjob": "{scheme}://{endpoint}/apis/kubeflow.org/v1alpha1/namespaces/{namespace}/tfjobs/{name}",  # noqa
    "list_tfjobs": "{scheme}://{endpoint}/apis/kubeflow.org/v1alpha1/namespaces/{namespace}/tfjobs",  # noqa
    "log": "{scheme}://{endpoint}/api/v1/namespaces/{namespace}/pods/{name}/log",
    "nodes": "{scheme}://{endpoint}/api/v1/nodes/",
    "configmaps": "{scheme}://{endpoint}/api/v1/namespaces/{namespace}/configmaps?{selector}",  # noqa
    "deployments": "{scheme}://{endpoint}/apis/apps/v1/namespaces/{namespace}/deployments/{name}",  # noqa
    "scale": "{scheme}://{endpoint}/apis/apps/v1/namespaces/{namespace}/deployments/{name}/scale",  # noqa
}

PUT_OPS = {
    "scale": (
        "{scheme}://{endpoint}/apis/apps/v1/namespaces/{namespace}/deployments/{name}/scale",
        "autoscaling/v1",
    )  # noqa
}

POST_OPS = {
    "namespace": ("{scheme}://{endpoint}/api/v1/namespaces", "v1"),
    "job": ("{scheme}://{endpoint}/apis/batch/v1/namespaces/{namespace}/jobs", "batch/v1"),  # noqa
    "tfjob": (
        "{scheme}://{endpoint}/apis/kubeflow.org/v1alpha1/namespaces/{namespace}/tfjobs",
        "kubeflow.org/v1alpha1",
    ),  # noqa
    "quota": ("{scheme}://{endpoint}/api/v1/namespaces/{namespace}/resourcequotas", "v1"),  # noqa
}

DELETE_OPS = {
    "namespace": ("{scheme}://{endpoint}/api/v1/namespaces/{namespace}", "v1"),
    "execution": ("{scheme}://{endpoint}/api/v1/namespaces/{namespace}/pods/{name}", "v1"),  # noqa
    "job": ("{scheme}://{endpoint}/apis/batch/v1/namespaces/{namespace}/jobs/{name}", "batch/v1"),  # noqa
    "tfjob": (
        "{scheme}://{endpoint}/apis/kubeflow.org/v1alpha1/namespaces/{namespace}/tfjobs/{name}",
        "v1",
    ),  # noqa
}


class K8SCaller(object):
    constants = const
    _type = "k8s"

    def __init__(self, ssl_context, endpoint, session, scheme="https"):
        self.ssl_context = ssl_context
        self.endpoint = endpoint
        self.session = session
        self.scheme = scheme

    @property
    def type(self):
        return self._type

    async def get_namespace(self, namespace):
        url = GET_OPS["namespace"]
        url = url.format(namespace=namespace, endpoint=self.endpoint, scheme=self.scheme)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8SNamespace(data=result)

    async def create_namespace(self, namespace):
        url, version = POST_OPS["namespace"]
        url = url.format(namespace=namespace, endpoint=self.endpoint, scheme=self.scheme)
        obj = K8SNamespace(namespace)
        return await self.post(url, version, obj.payload())

    async def delete_namespace(self, namespace):
        url, version = DELETE_OPS["namespace"]
        url = url.format(namespace=namespace, endpoint=self.endpoint, scheme=self.scheme)
        obj = K8SDelete()
        return await self.delete(url, version, obj.payload())

    async def wait_added(self, kind, namespace, name=None, timeout=30):
        url = WATCH_OPS[kind]
        url = url.format(namespace=namespace, name=name, endpoint=self.endpoint, scheme=self.scheme)
        return await self.watch(url, value="ADDED", timeout=timeout)

    async def wait_deleted(self, kind, namespace, name=None, timeout=30):
        url = WATCH_OPS[kind]
        url = url.format(namespace=namespace, name=name, endpoint=self.endpoint, scheme=self.scheme)
        return await self.watch(url, value="DELETED", timeout=timeout)

    async def define_quota(self, namespace, cpu_limit=None, mem_limit=None):
        url, version = POST_OPS["quota"]
        url = url.format(namespace=namespace, endpoint=self.endpoint, scheme=self.scheme)
        obj = K8SQuota(namespace=namespace, max_cpu=cpu_limit, max_memory=mem_limit)
        return await self.post(url, version, obj.payload())

    async def get_nodes(self):
        url = GET_OPS["nodes"]
        url = url.format(endpoint=self.endpoint, scheme=self.scheme)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8SNodeList(data=result)

    async def get_scale_deploy(self, namespace, name):
        url = GET_OPS["scale"]
        url = url.format(namespace=namespace, name=name, endpoint=self.endpoint, scheme=self.scheme)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return result["status"]["replicas"]

    async def set_scale_deploy(self, namespace, name, scale):
        url, version = PUT_OPS["scale"]
        url = url.format(namespace=namespace, name=name, endpoint=self.endpoint, scheme=self.scheme)

        obj = K8SScale(namespace, name, scale)

        return await self.put(url, version, obj.payload())

    async def get_job(self, namespace, name):
        url = GET_OPS["job"]
        url = url.format(namespace=namespace, name=name, endpoint=self.endpoint, scheme=self.scheme)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8SJob(data=result)

    async def get_tfjob(self, namespace, name):
        url = GET_OPS["tfjob"]
        url = url.format(namespace=namespace, name=name, endpoint=self.endpoint, scheme=self.scheme)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8STFJob(data=result)

    async def get_job_executions(self, namespace, name):
        url = GET_OPS["executions"]
        url = url.format(namespace=namespace, name=name, endpoint=self.endpoint, scheme=self.scheme)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8SExecutionList(data=result)

    async def get_execution_log(self, namespace, job_id, execution_id):
        url = GET_OPS["log"]
        url = url.format(namespace=namespace, name=execution_id, endpoint=self.endpoint, scheme=self.scheme)
        result = await self.get(url, noaccept=True)
        if result is None:
            return None
        else:
            return result

    async def get_execution_log_watch(self, namespace, job_id, execution_id):
        url = GET_OPS["log"]
        url = url.format(namespace=namespace, name=execution_id, endpoint=self.endpoint, scheme=self.scheme)
        async for logline in await self._watch_log(url, timeout=3660):
            yield logline

    async def delete_job(self, namespace, name, wait=False, purge=True):
        url, version = DELETE_OPS["job"]
        url = url.format(namespace=namespace, name=name, endpoint=self.endpoint, scheme=self.scheme)
        obj = K8SDelete(purge)
        await self.delete(url, version, obj.payload())
        if wait:
            return await self.wait_deleted("job", namespace, name)
        return True

    async def delete_execution(self, namespace, job_id, execution_id, wait=False, purge=True):
        url, version = DELETE_OPS["execution"]
        url = url.format(namespace=namespace, name=execution_id, endpoint=self.endpoint, scheme=self.scheme)
        obj = K8SDelete(purge)
        await self.delete(url, version, obj.payload())
        if wait:
            return await self.wait_deleted("execution", namespace, execution_id)
        return True

    async def delete_tfjob(self, namespace, name, wait=False):
        url, version = DELETE_OPS["tfjob"]
        url = url.format(namespace=namespace, name=name, endpoint=self.endpoint, scheme=self.scheme)
        obj = K8SDelete()
        await self.delete(url, version, obj.payload())
        if wait:
            return await self.wait_deleted("tfjob", namespace, name)
        return True

    async def list_jobs(self, namespace):
        url = GET_OPS["list_jobs"]
        url = url.format(namespace=namespace, endpoint=self.endpoint, scheme=self.scheme)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8SJobList(data=result)

    async def list_tfjobs(self, namespace):
        url = GET_OPS["list_tfjobs"]
        url = url.format(namespace=namespace, endpoint=self.endpoint, scheme=self.scheme)
        result = await self.get(url)
        if result is None:
            return None
        else:
            return K8STFJobList(data=result)

    async def create_job(
        self,
        namespace,
        name,
        image,
        command=None,
        args=None,
        cpu_limit=None,
        mem_limit=None,
        envvars={},
        volumes=None,
        volumeMounts=None,
        envFrom=None,
        entrypoint=None,
        **kw,
    ):
        url, version = POST_OPS["job"]
        url = url.format(namespace=namespace, name=name, endpoint=self.endpoint, scheme=self.scheme)
        obj = K8SJob(
            namespace=namespace,
            name=name,
            image=image,
            command=command,
            args=args,
            cpu_limit=cpu_limit,
            mem_limit=mem_limit,
            envvars=envvars,
            volumes=volumes,
            volumeMounts=volumeMounts,
            envFrom=envFrom,
            entrypoint=entrypoint,
            **kw,
        )
        return await self.post(url, version, obj.payload())

    async def create_tfjob(
        self,
        namespace,
        name,
        image,
        command=None,
        args=None,
        cpu_limit=None,
        mem_limit=None,
        envvars={},
        workers=1,
        ps=1,
        masters=1,
        tb_gs=None,
        volumes=None,
        volumeMounts=None,
        envFrom=None,
        entrypoint=None,
        **kw,
    ):
        url, version = POST_OPS["tfjob"]
        url = url.format(namespace=namespace, name=name, endpoint=self.endpoint, scheme=self.scheme)
        obj = K8STFJob(
            namespace=namespace,
            name=name,
            image=image,
            command=command,
            args=args,
            cpu_limit=cpu_limit,
            mem_limit=mem_limit,
            envvars=envvars,
            workers=workers,
            ps=ps,
            masters=masters,
            tb_gs=tb_gs,
            volumes=volumes,
            volumeMounts=volumeMounts,
            envFrom=envFrom,
            entrypoint=entrypoint,
            **kw,
        )
        return await self.post(url, version, obj.payload())

    async def wait_for_job(self, namespace, name):
        url = WATCH_OPS["job"]
        url = url.format(namespace=namespace, name=name, endpoint=self.endpoint, scheme=self.scheme)
        try:
            # We only wait 1 hour
            async for data in self._watch(url, 3660):
                json_data = json.loads(data)
                if "conditions" not in json_data["object"]["status"]:
                    continue
                for condition in json_data["object"]["status"]["conditions"]:
                    if condition["type"] == "Complete":
                        return json_data["object"]["status"]["succeeded"]
        except concurrent.futures._base.TimeoutError:
            return None

    # BASIC OPS

    async def _watch(self, url, timeout=20):
        async with self.session.get(url + "?watch=true", ssl=self.ssl_context, timeout=timeout) as resp:
            assert resp.status == 200
            while True:
                data = await resp.content.readline()
                yield data

    async def _watch_log(self, url, timeout=20):
        async with self.session.get(url + "?follow=true", ssl=self.ssl_context, timeout=timeout) as resp:
            assert resp.status == 200
            while True:
                data = await resp.content.readline()
                yield data

    async def watch(self, url, value=None, timeout=20):
        try:
            async for data in self._watch(url, timeout):
                json_data = json.loads(data)
                if json_data["type"] == value:
                    break
        except concurrent.futures._base.TimeoutError:
            pass

    async def get(self, url, noaccept=False):
        headers = {}
        if noaccept is False:
            headers["Accept"] = "application/json"
        async with self.session.get(url, headers=headers, ssl=self.ssl_context) as resp:
            if resp.status == 404:
                return None
            if resp.status != 200:
                text = await resp.text()
                logger.error("Error: %d - %s" % (resp.status, text))
            assert resp.status == 200
            if noaccept:
                return await resp.text()
            else:
                return await resp.json()

    async def patch(self, url, payload):
        async with self.session.patch(
            url,
            json=payload,
            headers={"Accept": "application/json", "Content-Type": "application/json-patch+json"},
            ssl=self.ssl_context,
        ) as resp:
            if resp.status in [201, 200]:
                return await resp.json()
            else:
                text = await resp.text()
                raise Exception("Error calling k8s: %d - %s" % (resp.status, text))

    async def put(self, url, version, payload):
        payload["apiVersion"] = version
        async with self.session.put(
            url, json=payload, headers={"Accept": "application/json"}, ssl=self.ssl_context
        ) as resp:
            if resp.status in [201, 200]:
                return await resp.json()
            else:
                text = await resp.text()
                raise Exception("Error calling k8s: %d - %s" % (resp.status, text))

    async def post(self, url, version, payload):
        payload["apiVersion"] = version
        async with self.session.post(
            url, json=payload, headers={"Accept": "application/json"}, ssl=self.ssl_context
        ) as resp:
            if resp.status in [201, 200]:
                return await resp.json()
            else:
                text = await resp.text()
                raise Exception("Error calling k8s: %d - %s" % (resp.status, text))

    async def delete(self, url, version, payload):
        payload["apiVersion"] = version
        async with self.session.delete(
            url, headers={"Accept": "application/json"}, json=payload, ssl=self.ssl_context
        ) as resp:
            if resp.status == 200:
                return await resp.json()
            elif resp.status == 404:
                return None
            elif resp.status == 409:
                return None
            else:
                text = await resp.text()
                raise Exception("Error calling k8s: %d - %s" % (resp.status, text))

    async def get_config_maps(self, namespace, labels=None):
        selector = ""
        if labels:
            selector = "labelSelector={}".format(",".join(f"{k}={v}" for k, v in labels.items()))
        url = GET_OPS["configmaps"]
        url = url.format(namespace=namespace, endpoint=self.endpoint, selector=selector, scheme=self.scheme)
        return await self.get(url)



import pytest

pytestmark = pytest.mark.asyncio


IPV6_TEMPLATE = [{
    "ChangeMode": "restart",
    "ChangeSignal": "",
    "DestPath": "/aio-environment/resolv.conf",
    "EmbeddedTmpl": "nameserver ::1\n",
    "LeftDelim": "{{",
    "Perms": "0644",
    "RightDelim": "}}",
    "SourcePath": "",
    "Splay": 5000000000
}]

IPV6_VOLUMES = [
    "aio-environment/resolv.conf:/etc/resolv.conf"
]


async def test_get_jobs_nomad(nomad):
    # We clean up all the jobs on the namespace
    result = await nomad.get_nodes()
    assert len(result) > 0

    result = await nomad.cleanup_jobs('aiocluster-test')
    assert result == 0

    jobs_info = await nomad.list_jobs('aiocluster-test')
    assert jobs_info.total == 0

    await nomad.create_job(
        'aiocluster-test',
        'test-job',
        'perl',
        command=["perl"],
        args=["-Mbignum=bpi", "-wle", "print bpi(2000)"],
        timeout=45,
        delete=True,
        wait=False,
        templates=IPV6_TEMPLATE,
        volumes=IPV6_VOLUMES)

    job_info = await nomad.get_job('aiocluster-test', 'test-job')
    assert job_info.id == 'test-job'

    assert job_info._raw['TaskGroups'][0]['Tasks'][0]['Config']['volumes'] == IPV6_VOLUMES  # noqa

    jobs_info = await nomad.list_jobs('aiocluster-test')
    assert jobs_info.total == 1

    result = await nomad.wait_for_job('aiocluster-test', 'test-job')
    assert result == 1

    job_info = await nomad.get_job('aiocluster-test', 'test-job')
    assert job_info.finished

    job_info = await nomad.get_job('aiocluster-test', 'test-job')
    assert job_info.id == 'test-job'

    scale = await nomad.get_scale_deploy('aiocluster-test', 'test-job')
    assert scale == 1

    await nomad.set_scale_deploy('aiocluster-test', 'test-job', 2)

    scale = await nomad.get_scale_deploy('aiocluster-test', 'test-job')
    assert scale == 2

    executions = await nomad.list_job_executions(
        'aiocluster-test', 'test-job')
    assert len(executions) > 0

    # Testing does not work on local nomad
    # log = await nomad.get_execution_log(
    #     'aiocluster-test', 'test-job', executions[0].internal_id)
    # assert "3.14" in log

    result = await nomad.delete_job('aiocluster-test', 'test-job')
    assert result is True

    job_info = await nomad.get_job('aiocluster-test', 'test-job')
    assert job_info is None
